"""
.. include:: ../../README.md
"""

from cstag.call import call
from cstag.shorten import shorten
from cstag.lengthen import lengthen
from cstag.consensus import consensus
from cstag.mask import mask
from cstag.split import split
from cstag.revcomp import revcomp
from cstag.to_html import to_html
from cstag.to_pdf import to_pdf
from cstag.to_vcf import to_vcf
from cstag.to_sequence import to_sequence
